import time
import requests
import os
import logging


logging.basicConfig(level=logging.WARNING)


def delete_old_database():
    dirname = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
    filename = 'example_project/example_project/db.sqlite3'
    path = os.path.abspath(os.path.join(dirname, filename))

    try:
        os.remove(path)
        logging.debug("Database file in example_project DELETED!")
    except OSError:
        logging.debug("No database file found.")
        pass


def obtain_csrftoken(session):
    """
    A CSRF cookie token is required in order to not get a 403 Forbidden response by
    the post to the inputs.
    :return: String representing the token.
    """

    return session.cookies['csrftoken']


def create_session():
    """
    A integration test utility to create a browser session request for a single test.
    :return: A session object for requests.
    """

    return requests.Session()


def send_get_request(session, url):
    """
    Attempts to send a GET to the url. Server is already up as the test should
    be calling the `is_server_healthy()` function before calling this.

    :param session: Object representing the browser session.
    :param url: String, containing 'http://', with the URL of the target URL.
    :return: Response of GET.
    """
    response = None

    logging.debug("Attempting a GET request for the provided URL...")
    try:
        response = session.get(url)
        response.raise_for_status()
    except requests.exceptions.RequestException as e:
        logging.debug(e)

    return response


def send_post_request(session, url, data):
    """
    Attempts to send a POST to the url. Server is already up as the test should
    be calling the `is_server_healthy()` function before calling this.

    :return: Response of POST.
    """
    response = None

    logging.debug("Attempting a POST request for the provided URL...")

    try:
        response = session.post(url, data=data)
        response.raise_for_status()
    except requests.exceptions.RequestException as e:
        logging.debug(e)

    return response


def is_server_healthy(url):
    """
    Function will only return True when the param URL returns a 2xx code. After
    45 seconds, the check assumes a timeout.
    :param url: http URL for the address to poll.
    :return: boolean value to indicate result.
    """

    attempts = 0

    logging.debug("Checking if the server is healthy...")
    while attempts <= 45:
        try:
            status_code = requests.get(url).status_code
            if int(str(status_code)[0]) == 2:
                return True
        except requests.exceptions.RequestException as e:
            pass

        attempts += 1
        time.sleep(1)
    return False


def _log_in_as_a_superuser():
    """
    A private wrapper function for all the utilities that will
    log a user in with the correct credentials and take care of
    all CSRF token exchange.
    """
    url = 'http://localhost:8000/kurono/accounts/login/'
    assert(is_server_healthy(url))

    logging.debug("Creating session...")
    session = create_session()

    send_get_request(session, url)

    logging.debug("Obtaining CSRF Token...")
    csrftoken = obtain_csrftoken(session)

    login_info = {
        'username': 'admin',
        'password': 'admin',
        'csrfmiddlewaretoken': csrftoken,
    }

    logging.debug("Sending post response...")

    response = send_post_request(session, url, login_info)
    assert(response.status_code == 200)

    return csrftoken, session


def create_custom_game_default_settings(name):
    """
    Sends an appropriate POST request to create a game with a
    given name, using default settings provided.
    """
    csrftoken, session = _log_in_as_a_superuser()

    url = 'http://localhost:8000/kurono/games/new/'

    print("is server healthy? ", is_server_healthy(url))

    csrftoken = session.cookies['csrftoken']

    data = {
        "csrfmiddlewaretoken": csrftoken,
        "name": name,
        "public": "on",
        "can_play": "1",
        "generator": "Main",
        "target_num_cells_per_avatar": "16",
        "target_num_score_locations_per_avatar": "0.5",
        "score_despawn_chance": "0.02",
        "target_num_pickups_per_avatar": "0.5",
        "pickup_spawn_chance": "0.02",
        "obstacle_ratio": "0.1",
        "start_height": "11",
        "start_width": "11",
    }

    headers = {'X-CSRFToken': csrftoken, 'Referer': 'http://localhost:8000/kurono/accounts/login/'}

    response = session.post(url, data=data, headers=headers)

    return response



from __future__ import absolute_import
from __future__ import print_function

import json
import os
import platform
import re

import docker

from .shell_api import BASE_DIR, run_command


def create_docker_client(use_raw_env=False, minikube=None):
    """
    Create a docker client using the python SDK.

    :param raw_env_settings: String that is returned by the 'minikube docker-env' command.
    :return:
    """
    if use_raw_env:
        raw_env_settings = run_command(
            [minikube, "docker-env", '--shell="bash"'], True
        ).decode("utf-8")
        matches = re.finditer(r'^export (.+)="(.+)"$', raw_env_settings, re.MULTILINE)
        env_variables = dict([(m.group(1), m.group(2)) for m in matches])

    else:
        # VM driver is set
        env_variables = os.environ

    env_variables["DOCKER_BUILDKIT"] = "1"
    return docker.from_env(environment=env_variables, version="auto")


def build_docker_images(minikube=None, build_target=None):
    """
    Find environment settings and builds docker images for each directory.

    :param minikube: Executable command to run in terminal.
    """
    print("Building docker images")
    if minikube:
        client = create_docker_client(use_raw_env=True, minikube=minikube)
    else:
        client = create_docker_client(use_raw_env=False, minikube=minikube)

    directories = ("aimmo-game", "aimmo-game-creator", "aimmo-game-worker")
    for dir in directories:
        path = os.path.join(BASE_DIR, dir)
        tag = "ocadotechnology/%s:test" % dir
        print("Building %s..." % tag)
        client.images.build(path=path, tag=tag, encoding="gzip", target=build_target)


def delete_containers():
    """Delete any containers starting with 'aimmo'."""
    client = docker.from_env(version="auto")

    containers = [
        container
        for container in client.containers.list(all=True)
        if container.name.startswith("aimmo")
    ]
    for container in containers:
        container.remove(force=True)


def start_game_creator():
    """Start an aimmo-game-creator docker container."""
    os_name = platform.system()
    client = docker.from_env(version="auto")
    template = {
        "detach": True,
        "tty": True,
        "environment": {"FLASK_ENV": "development", "WORKER": "local"},
        "volumes": {
            "/var/run/docker.sock": {"bind": "/var/run/docker.sock", "mode": "rw"},
            "aimmo_game_tokens": {"bind": "/tokens", "mode": "rw"},
        },
    }
    if os_name == "Linux":
        template["environment"]["LOCALHOST_IP"] = "127.0.0.1"
        template["network_mode"] = "host"
    else:
        template["environment"]["LOCALHOST_IP"] = "host.docker.internal"

    template["environment"]["CONTAINER_TEMPLATE"] = json.dumps(template)
    kwargs = template.copy()
    client.containers.run(
        name="aimmo-game-creator",
        image="ocadotechnology/aimmo-game-creator:test",
        **kwargs
    )



from abc import ABCMeta, abstractmethod, abstractproperty
from typing import TYPE_CHECKING, List

from simulation.interactables.conditions import TurnState

if TYPE_CHECKING:
    from simulation.game_state import GameState
    from simulation.avatar.avatar_wrapper import AvatarWrapper


class _Interactable(object):
    __metaclass__ = ABCMeta

    def __init__(self, cell):
        self.cell = cell
        self.delete_after_effects_applied = False

        self.conditions = []
        self.effects = []
        self.targets = []

    def __str__(self):
        return self.__class__.__name__

    def delete(self):
        self.cell.interactable = None

    def conditions_met(self, game_state: "GameState") -> "bool":
        """Apply logical `AND` on all conditions, returns True if all conditions are met."""
        turn_state = TurnState(game_state, self.cell)
        return all([condition(turn_state) for condition in self.conditions])

    def apply(self):
        """
        Apply all effects in sequential order.

        Targets for effects can be a single object, or a list of objects. All
        targets must have an 'effect' attribute that is of type=set.
        """
        self.targets = self.get_targets()
        for effect in self.effects:
            for target in self.targets:
                effect(target)

        if self.delete_after_effects_applied:
            self.delete()

    @abstractmethod
    def serialize(self):
        raise NotImplementedError()

    @abstractmethod
    def get_targets(self) -> "List[AvatarWrapper]":
        raise NotImplementedError()



from __future__ import absolute_import
from .base_test_migration import MigrationTestCase


class TestMigrationAlterGameToken(MigrationTestCase):

    start_migration = "0009_add_game_status"
    dest_migration = "0010_alter_game_token"

    def test_game_token_field_altered(self):
        model = self.django_application.get_model(self.app_name, "Game")
        game = model.objects.create(id=1, name="test", public=True)

        self.assertEquals(game.auth_token, "")



from __future__ import absolute_import

import unittest

from simulation import action
from simulation import event
from simulation.avatar.avatar_manager import AvatarManager
from simulation.direction import EAST
from simulation.game_state import GameState
from simulation.location import Location
from .dummy_avatar import MoveDummy
from .maps import InfiniteMap, EmptyMap, AvatarMap

ORIGIN = Location(x=0, y=0)
EAST_OF_ORIGIN = Location(x=1, y=0)
NORTH_OF_ORIGIN = Location(x=0, y=1)


class TestAction(unittest.TestCase):
    def setUp(self):
        self.avatar = MoveDummy(1, ORIGIN, EAST)
        self.other_avatar = MoveDummy(2, EAST_OF_ORIGIN, EAST)
        self.avatar_manager = AvatarManager()

    def test_successful_move_action(self):
        # Move north
        game_state = GameState(InfiniteMap(), self.avatar_manager)
        action.MoveAction(self.avatar, {"x": 0, "y": 1}).process(game_state.world_map)

        target_cell = game_state.world_map.get_cell(NORTH_OF_ORIGIN)
        self.assertEqual(self.avatar.location, NORTH_OF_ORIGIN)
        self.assertEqual(self.avatar, target_cell.avatar)

        self.assertEqual(
            self.avatar.events, [event.MovedEvent(ORIGIN, NORTH_OF_ORIGIN)]
        )

        # Move east
        self.setUp()
        game_state = GameState(InfiniteMap(), self.avatar_manager)
        action.MoveAction(self.avatar, {"x": 1, "y": 0}).process(game_state.world_map)

        self.assertEqual(self.avatar.location, EAST_OF_ORIGIN)
        self.assertEqual(self.avatar.events, [event.MovedEvent(ORIGIN, EAST_OF_ORIGIN)])

    def test_successful_move_east_twice_action(self):
        game_state = GameState(InfiniteMap(), self.avatar_manager)
        action.MoveAction(self.avatar, {"x": 1, "y": 0}).process(game_state.world_map)
        action.MoveAction(self.avatar, {"x": 1, "y": 0}).process(game_state.world_map)

        self.assertEqual(self.avatar.location, Location(2, 0))

    def test_failed_move_action(self):
        game_state = GameState(EmptyMap(), self.avatar_manager)
        action.MoveAction(self.avatar, {"x": 0, "y": 1}).process(game_state.world_map)

        self.assertEqual(self.avatar.location, ORIGIN)
        self.assertEqual(
            self.avatar.events, [event.FailedMoveEvent(ORIGIN, NORTH_OF_ORIGIN)]
        )

    def test_successful_attack_action(self):
        game_state = GameState(AvatarMap(self.other_avatar), self.avatar_manager)
        action.AttackAction(self.avatar, {"x": 0, "y": 1}).process(game_state.world_map)

        target_location = NORTH_OF_ORIGIN
        damage_dealt = 1

        self.assertEqual(self.avatar.location, ORIGIN)
        self.assertEqual(self.other_avatar.location, EAST_OF_ORIGIN)
        self.assertEqual(self.other_avatar.times_died, 0)
        self.assertEqual(self.other_avatar.health, 4)

        self.assertEqual(
            self.avatar.events,
            [
                event.PerformedAttackEvent(
                    self.other_avatar, target_location, damage_dealt
                )
            ],
        )
        self.assertEqual(
            self.other_avatar.events,
            [event.ReceivedAttackEvent(self.avatar, damage_dealt)],
        )

    def test_successful_multiple_attack_actions(self):
        game_state = GameState(AvatarMap(self.other_avatar), self.avatar_manager)
        action.AttackAction(self.avatar, {"x": 0, "y": 1}).process(game_state.world_map)

        self.assertEqual(
            self.other_avatar.events, [event.ReceivedAttackEvent(self.avatar, 1)]
        )

        action.AttackAction(self.avatar, {"x": 0, "y": 1}).process(game_state.world_map)

        self.assertEqual(
            self.other_avatar.events,
            [
                event.ReceivedAttackEvent(self.avatar, 1),
                event.ReceivedAttackEvent(self.avatar, 1),
            ],
        )

        self.assertEqual(self.avatar.location, ORIGIN)
        self.assertEqual(self.other_avatar.location, EAST_OF_ORIGIN)
        self.assertEqual(self.other_avatar.times_died, 0)
        self.assertEqual(self.other_avatar.health, 3)

    def test_failed_attack_action(self):
        game_state = GameState(InfiniteMap(), self.avatar_manager)
        action.AttackAction(self.avatar, {"x": 0, "y": 1}).process(game_state.world_map)

        target_location = NORTH_OF_ORIGIN

        self.assertEqual(self.avatar.location, ORIGIN)
        self.assertEqual(self.other_avatar.location, EAST_OF_ORIGIN)
        self.assertEqual(self.avatar.events, [event.FailedAttackEvent(target_location)])
        self.assertEqual(self.other_avatar.events, [])

    def test_avatar_dies(self):
        self.other_avatar.health = 1
        game_state = GameState(AvatarMap(self.other_avatar), self.avatar_manager)
        action.AttackAction(self.avatar, {"x": 0, "y": 1}).process(game_state.world_map)

        target_location = NORTH_OF_ORIGIN
        damage_dealt = 1
        self.assertEqual(
            self.avatar.events,
            [
                event.PerformedAttackEvent(
                    self.other_avatar, target_location, damage_dealt
                )
            ],
        )
        self.assertEqual(
            self.other_avatar.events,
            [event.ReceivedAttackEvent(self.avatar, damage_dealt)],
        )

        self.assertEqual(self.avatar.location, ORIGIN)
        self.assertEqual(self.other_avatar.health, 0)
        self.assertEqual(self.other_avatar.times_died, 1)
        self.assertEqual(self.other_avatar.location, Location(10, 10))

    def test_no_move_in_wait(self):
        game_state = GameState(InfiniteMap(), self.avatar_manager)
        action.WaitAction(self.avatar).process(game_state.world_map)
        self.assertEqual(self.avatar.location, ORIGIN)



from simulation.avatar.avatar_wrapper import AvatarWrapper
from simulation.avatar.avatar_appearance import AvatarAppearance


class AvatarManager(object):
    """
    Stores all game avatars.
    """

    def __init__(self):
        self.avatars_by_id = {}

    def add_avatar(self, player_id, location):
        avatar = AvatarWrapper(
            player_id, location, AvatarAppearance("#000", "#ddd", "#777", "#fff")
        )
        self.avatars_by_id[player_id] = avatar
        return avatar

    def get_avatar(self, user_id):
        return self.avatars_by_id[user_id]

    def remove_avatar(self, user_id):
        del self.avatars_by_id[user_id]

    @property
    def avatars(self):
        return list(self.avatars_by_id.values())

    @property
    def active_avatars(self):
        return self.avatars[:]

    def serialize_players(self):
        """
        To be called on a required update of players in the front-end of the game.
        """

        return [player.serialize() for player in self.avatars]



from __future__ import absolute_import

import logging
from simulation.action import MoveAction, WaitAction
from simulation.avatar.avatar_manager import AvatarManager
from simulation.avatar.avatar_wrapper import AvatarWrapper
from simulation.direction import NORTH, EAST, SOUTH, WEST

LOGGER = logging.getLogger(__name__)


class DummyAvatar(AvatarWrapper):
    def __init__(self, player_id=1, initial_location=(0, 0)):
        # TODO: extract avatar state and state-altering methods into a new class.
        #       The new class is to be shared between DummyAvatarRunner and AvatarRunner
        super(DummyAvatar, self).__init__(
            player_id, initial_location, avatar_appearance=None
        )
        self.times_died = 0
        self.attack_strength = 1
        self.effects = set()
        self.resistance = 0

    def decide_action(self, worker_data):
        raise NotImplementedError()

    def next_turn(self, world_map=None, avatar_state=None):
        raise NotImplementedError()

    def add_event(self, event):
        self.events.append(event)

    def die(self, respawn_loc):
        self.location = respawn_loc
        self.times_died += 1

    def fetch_data(self, state_view):
        return {"action": "", "log": "Testing", "avatar_updated": False}

    def damage(self, amount):
        self.health -= amount
        return amount


class LiveDummy(DummyAvatar):
    """
    An avatar that still has a fully functioning worker.
    """

    def __init__(self, player_id=1, initial_location=(0, 0)):
        super(LiveDummy, self).__init__(player_id, initial_location)

    def decide_action(self, worker_data):
        self._action = self.next_turn()
        return True


class DeadDummy(DummyAvatar):
    """
    An avatar whose worker is no longer responding or returning an invalid action.
    """

    def __init__(self, player_id=1, initial_location=(0, 0)):
        super(DeadDummy, self).__init__(player_id, initial_location)

    def decide_action(self, worker_data):
        self._action = self.next_turn()
        return False

    def next_turn(self, world_map=None, avatar_state=None):
        return WaitAction(self)


class WaitDummy(LiveDummy):
    """
    Avatar that always waits.
    """

    def next_turn(self, world_map=None, avatar_state=None):
        return WaitAction(self)


class MoveDummy(LiveDummy):
    """
    Avatar that always moves in one direction.
    """

    def __init__(self, player_id, initial_location, direction):
        super(MoveDummy, self).__init__(player_id, initial_location)
        self._direction = direction

    def next_turn(self, world_map=None, avatar_state=None):
        return MoveAction(self, self._direction.dict)


class MoveNorthDummy(MoveDummy):
    def __init__(self, player_id, initial_location):
        super(MoveNorthDummy, self).__init__(player_id, initial_location, NORTH)


class MoveEastDummy(MoveDummy):
    def __init__(self, player_id, initial_location):
        super(MoveEastDummy, self).__init__(player_id, initial_location, EAST)


class MoveSouthDummy(MoveDummy):
    def __init__(self, player_id, initial_location):
        super(MoveSouthDummy, self).__init__(player_id, initial_location, SOUTH)


class MoveWestDummy(MoveDummy):
    def __init__(self, player_id, initial_location):
        super(MoveWestDummy, self).__init__(player_id, initial_location, WEST)


class DummyAvatarManager(AvatarManager):
    def __init__(self, dummy_list=None):
        super(DummyAvatarManager, self).__init__()
        if dummy_list is None:
            dummy_list = []
        self.dummy_list = dummy_list

    def add_avatar(self, player_id, location=(0, 0)):
        try:
            dummy = self.dummy_list.pop(0)
        except IndexError:
            dummy = WaitDummy
        self.avatars_by_id[player_id] = dummy(player_id, location)
        return self.avatars_by_id[player_id]

    def add_avatar_directly(self, avatar):
        self.avatars_by_id[avatar.player_id] = avatar

    def get_player_id_to_serialized_action(self):
        for dummy in self.avatars_by_id.values():
            dummy.decide_action(None)

        return {
            player_id: self.avatars_by_id[player_id]._action
            for player_id in self.avatars_by_id
        }



import logging
import mock
from unittest import TestCase
from aimmo_runner.minikube import create_creator_yaml

logging.basicConfig(level=logging.WARNING)


class TestMinikubeRunner(TestCase):

    @mock.patch('aimmo_runner.minikube.get_ip', return_value='127.0.0.1')
    def test_game_creator_function_creates_correct_game_url(self, mocked_get_ip_func):
        """
        Checks if the game creator yaml is created with the correct game API URL. Relies on get_ip()
        which we expect to fall back to localhost always.
        """
        created_yaml = create_creator_yaml()
        self.assertEqual(created_yaml['spec']['template']['spec']['containers'][0]['env'][1]['value'],
                         'http://127.0.0.1:8000/kurono/api/games/')



"""
Interactables is a package for adding objects to the world map.

This contains the definitions of these objects, including all their effects and conditions,
multiple interactables of the same type are grouped (pickups for example).
"""


def serialize_interactables(world_map):
    return []



import os

import aiohttp
import requests


class DjangoCommunicator(object):
    """
    This class encapsulates the communication between aimmo-game
    and the django server
    """

    def __init__(self, django_api_url, completion_url):
        self.session = aiohttp.ClientSession()
        self.django_api_url = django_api_url
        self.completion_url = completion_url
        self.token_url = self.django_api_url + "token/"

    def get_game_metadata(self):
        return requests.get(f"{self.django_api_url}users/").json()

    def mark_game_complete(self, data=None):
        return requests.post(requests.post(self.completion_url, json=data))

    async def patch_token(self, data):
        response = await self.session.patch(
            self.token_url, headers={"Game-token": os.environ["TOKEN"]}, json=data
        )
        return response

    async def patch_game(self, data):
        response = await self.session.patch(
            self.django_api_url, headers={"Game-token": os.environ["TOKEN"]}, json=data
        )
        return response

    async def close_session(self, app):
        await self.session.close()



from django.apps import AppConfig
from django.db.models.signals import post_migrate


def clear_tokens(sender, **kwargs):
    game_model = sender.get_model("Game")
    games = game_model.objects.all()
    games.update(auth_token="")


class AimmoAppConfig(AppConfig):
    name = "aimmo"

    def ready(self):
        post_migrate.connect(clear_tokens, sender=self)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("aimmo", "0003_auto_20160802_1418")]

    operations = [
        migrations.AddField(
            model_name="game",
            name="obstacle_ratio",
            field=models.FloatField(default=0.1),
        ),
        migrations.AddField(
            model_name="game",
            name="pickup_spawn_chance",
            field=models.FloatField(default=0.02),
        ),
        migrations.AddField(
            model_name="game",
            name="score_despawn_chance",
            field=models.FloatField(default=0.02),
        ),
        migrations.AddField(
            model_name="game",
            name="start_height",
            field=models.IntegerField(default=11),
        ),
        migrations.AddField(
            model_name="game", name="start_width", field=models.IntegerField(default=11)
        ),
        migrations.AddField(
            model_name="game",
            name="target_num_cells_per_avatar",
            field=models.FloatField(default=16),
        ),
        migrations.AddField(
            model_name="game",
            name="target_num_pickups_per_avatar",
            field=models.FloatField(default=0.5),
        ),
        migrations.AddField(
            model_name="game",
            name="target_num_score_locations_per_avatar",
            field=models.FloatField(default=0.5),
        ),
    ]



import logging
import os
import time

from kubernetes import client
from kubernetes import config

from .worker import Worker

LOGGER = logging.getLogger(__name__)
# Default here to stop import errors if imported when running locally
K8S_NAMESPACE = os.environ.get("K8S_NAMESPACE", "")


class KubernetesWorker(Worker):
    """Kubernetes worker."""

    def __init__(self, *args, **kwargs):
        config.load_incluster_config()
        self.api = client.CoreV1Api()
        self.game_id = os.environ["GAME_ID"]
        self.game_url = os.environ["GAME_URL"]
        self.pod_name = os.environ["POD_NAME"]
        super(KubernetesWorker, self).__init__(*args, **kwargs)

    @staticmethod
    def _create_a_label_selector_from_labels(label_list):
        return ",".join(label_list)

    def _make_owner_references(self):
        try:
            return [
                client.V1OwnerReference(
                    api_version="v1",
                    block_owner_deletion=True,
                    controller=True,
                    kind="Pod",
                    name=self.pod_name,
                    uid=self._get_game_uid(),
                )
            ]
        except IndexError:
            # Couldn't find the current pod
            return []

    def _make_container(self, player_id):
        return client.V1Container(
            env=[
                client.V1EnvVar(
                    name="DATA_URL", value="%s/player/%d" % (self.game_url, player_id)
                ),
                client.V1EnvVar(name="PORT", value="5000"),
            ],
            name="aimmo-game-worker",
            image="ocadotechnology/aimmo-game-worker:%s"
            % os.environ.get("IMAGE_SUFFIX", "latest"),
            ports=[client.V1ContainerPort(container_port=5000, protocol="TCP")],
            resources=client.V1ResourceRequirements(
                limits={"cpu": "10m", "memory": "64Mi"},
                requests={"cpu": "6m", "memory": "32Mi"},
            ),
            security_context=client.V1SecurityContext(
                capabilities=client.V1Capabilities(
                    drop=["all"], add=["NET_BIND_SERVICE"]
                )
            ),
        )

    def make_pod(self):
        pod_manifest = client.V1PodSpec(
            containers=[self._make_container(self.player_id)],
            service_account_name="worker",
        )

        metadata = client.V1ObjectMeta(
            labels={
                "app": "aimmo-game-worker",
                "game": self.game_id,
                "player": str(self.player_id),
            },
            generate_name="aimmo-%s-worker-%s-" % (self.game_id, self.player_id),
            owner_references=self._make_owner_references(),
        )
        return client.V1Pod(metadata=metadata, spec=pod_manifest)

    def _get_game_uid(self):
        pod_list = self.api.list_namespaced_pod(
            namespace=K8S_NAMESPACE, field_selector=f"metadata.name={self.pod_name}"
        )
        pod_metadata = pod_list.items[0].metadata
        return pod_metadata.uid

    def _wait_for_pod_creation(self, pod_name, player_id):

        for _ in range(90):
            pod = self.api.read_namespaced_pod(pod_name, K8S_NAMESPACE)
            LOGGER.info("Pod status: {}".format(pod.status))
            if pod.status.phase == "Running":
                return pod

            time.sleep(1)

        raise EnvironmentError("Could not start worker %s." % player_id)

    def _create_worker(self):
        pod_obj = self.make_pod()
        LOGGER.info("Making new workers pod: {}".format(pod_obj.metadata.name))
        pod = self.api.create_namespaced_pod(namespace=K8S_NAMESPACE, body=pod_obj)
        pod_name = pod.metadata.name
        pod = self._wait_for_pod_creation(pod_name, self.player_id)

        worker_url = "http://%s:5000" % pod.status.pod_ip
        LOGGER.info("Worker ip: {}".format(pod.status.pod_ip))
        LOGGER.info(
            "Worker started for %s, listening at %s", self.player_id, worker_url
        )
        return worker_url

    def remove_worker(self):
        app_label = "app=aimmo-game-worker"
        game_label = "game={}".format(self.game_id)
        player_label = "player={}".format(self.player_id)
        label_selector = self._create_a_label_selector_from_labels(
            [app_label, game_label, player_label]
        )
        pods = self.api.list_namespaced_pod(
            namespace=K8S_NAMESPACE, label_selector=label_selector
        )

        for pod in pods.items:
            LOGGER.info("Deleting pod: {}".format(pod.metadata.name))
            self.api.delete_namespaced_pod(
                pod.metadata.name, K8S_NAMESPACE, client.V1DeleteOptions()
            )



from aimmo.models import Game
from django.core.exceptions import ValidationError
from django.forms import ModelForm


class AddGameForm(ModelForm):
    def __init__(self, playable_games, *args, **kwargs):
        self.playable_games = playable_games
        super(AddGameForm, self).__init__(*args, **kwargs)

    class Meta:
        model = Game
        exclude = [
            "Main",
            "owner",
            "auth_token",
            "completed",
            "main_user",
            "static_data",
            "can_play",
            "public",
            "generator",
            "target_num_cells_per_avatar",
            "target_num_score_locations_per_avatar",
            "score_despawn_chance",
            "target_num_pickups_per_avatar",
            "pickup_spawn_chance",
            "obstacle_ratio",
            "start_height",
            "start_width",
            "status",
        ]

    def clean(self):
        name = self.cleaned_data["name"]

        playable_games_names = [
            playable_game.name
            for playable_game in self.playable_games
            if self.playable_games and name
        ]

        if name in playable_games_names:
            raise ValidationError("Sorry, a game with this name already exists.")

        return self.cleaned_data



import asyncio

import pytest
from simulation.interactables.score_location import ScoreLocation
from simulation.location import Location

from .mock_world import MockWorld

game = None
cell = None


@pytest.fixture
def test_data():
    """
    Mock a game for each test individually. MockWorld() will set up a:
    avatar manager, game state, turn manager and a map generator.
    """
    game = MockWorld()
    game.simulation_runner.add_avatar(1, Location(0, 0))
    cell = game.game_state.world_map.get_cell(Location(0, 0))

    # Avatar will try to move North, so we force it to stay in place
    game.game_state.world_map.get_cell(Location(1, 0)).habitable = False

    return game, cell


async def test_score_location_increase_score_of_avatar(test_data, loop):
    """
    Avatar spawns at the origin (0,0) and should have a score of 0. Moves
    EAST to (1,0) and should automatically then receive an effect that will
    increase the avatars score.
    """
    game, cell = test_data
    cell.interactable = ScoreLocation(cell)
    assert game.avatar_manager.get_avatar(1).score is 0
    assert cell.interactable.serialize() == {
        "type": "score",
        "location": {"x": cell.location.x, "y": cell.location.y},
    }

    await game.simulation_runner.run_single_turn(
        game.avatar_manager.get_player_id_to_serialized_action()
    )

    assert cell.avatar is game.avatar_manager.get_avatar(1)
    assert cell.avatar.score is 1
    assert len(cell.avatar.effects) is 1


async def test_score_locations_persist_and_keep_giving_score_effects(test_data, loop):
    """
    Checks if score can be increased more than once. First moved from ORIGIN to 1,0 ->
    then is given some score, remains in place due to an obstacle, then next turn
    it's score should increase again.
    """
    game, cell = test_data
    loop = asyncio.get_event_loop()
    cell.interactable = ScoreLocation(cell)

    await game.simulation_runner.run_single_turn(
        game.avatar_manager.get_player_id_to_serialized_action()
    )

    assert cell.avatar is game.avatar_manager.get_avatar(1)
    assert cell.avatar.score is 1

    await game.simulation_runner.run_single_turn(
        game.avatar_manager.get_player_id_to_serialized_action()
    )

    assert cell.avatar.score is 2
    assert cell.avatar is game.avatar_manager.get_avatar(1)



import logging
from asyncio import CancelledError

from aiohttp import ClientSession, ClientResponseError, ServerDisconnectedError

LOGGER = logging.getLogger(__name__)


class Worker(object):
    def __init__(self, player_id, game_port):
        self.log = None
        self.player_id = player_id
        self.game_port = game_port
        self.code = None
        self.serialized_action = None
        self.has_code_updated = False
        self.url = self._create_worker()
        self.ready = False

    def _set_defaults(self):
        self.log = None
        self.serialized_action = None
        self.has_code_updated = False

    def _create_worker(self):
        raise NotImplementedError

    def remove_worker(self):
        raise NotImplementedError

    async def fetch_data(self, state_view):
        if self.code is None:
            self._set_defaults()
            return
        try:
            async with ClientSession(raise_for_status=True) as session:
                code_and_options = {"code": self.code, "options": {}, "state": None}
                data = {**state_view, **code_and_options}
                response = await session.post(f"{self.url}/turn/", json=data)
                data = await response.json()
                self.serialized_action = data["action"]
                self.log = data["log"]
                self.has_code_updated = data["avatar_updated"]
                self.ready = True
        except (ClientResponseError, ServerDisconnectedError):
            LOGGER.info("Could not connect to worker, probably not ready yet")
            self._set_defaults()
        except CancelledError as e:
            LOGGER.error("Worker took too long to respond: {}".format(e))
            self._set_defaults()
        except KeyError as e:
            LOGGER.error("Missing key in data from worker: {}".format(e))
            self._set_defaults()
        except Exception as e:
            LOGGER.exception("Unknown error while fetching turn data.")
            LOGGER.exception(e)
            self._set_defaults()



from __future__ import absolute_import

import random
import unittest

from simulation import map_generator
from simulation.location import Location
from simulation.map_generator import get_random_edge_index
from simulation.simulation_runner import SequentialSimulationRunner
from simulation.world_map import WorldMap

from .dummy_avatar import DummyAvatarManager


class ConstantRng(object):
    def __init__(self, value):
        self.value = value

    def randint(self, minimum, maximum):
        if not minimum <= self.value <= maximum:
            raise ValueError("Beyond range")
        return self.value


class TestHelperFunctions(unittest.TestCase):
    def test_get_random_edge_index(self):
        map = WorldMap.generate_empty_map(3, 4, {})
        self.assertEqual((0, -1), get_random_edge_index(map, rng=ConstantRng(0)))
        self.assertEqual((1, -1), get_random_edge_index(map, rng=ConstantRng(1)))
        self.assertEqual((0, 1), get_random_edge_index(map, rng=ConstantRng(2)))
        self.assertEqual((1, 1), get_random_edge_index(map, rng=ConstantRng(3)))
        self.assertEqual((-1, 0), get_random_edge_index(map, rng=ConstantRng(4)))
        self.assertEqual((2, 0), get_random_edge_index(map, rng=ConstantRng(5)))

        # Verify no out of bounds
        with self.assertRaisesRegex(ValueError, "Beyond range"):
            get_random_edge_index(map, rng=ConstantRng(-1))

        with self.assertRaisesRegex(ValueError, "Beyond range"):
            get_random_edge_index(map, rng=ConstantRng(6))

    def test_get_random_edge_index_can_give_all_possible(self):
        map = WorldMap.generate_empty_map(3, 4, {})
        get_random_edge_index(map, rng=ConstantRng(1))
        expected = frozenset(((0, 1), (1, 1), (-1, 0), (2, 0), (0, -1), (1, -1)))
        actual = frozenset(
            get_random_edge_index(map, rng=ConstantRng(i)) for i in range(6)
        )
        self.assertEqual(expected, actual)

    def test_out_of_bounds_random_edge(self):
        map = WorldMap.generate_empty_map(3, 4, {})
        with self.assertRaisesRegex(ValueError, "Beyond range"):
            get_random_edge_index(map, rng=ConstantRng(-1))

        with self.assertRaisesRegex(ValueError, "Beyond range"):
            get_random_edge_index(map, rng=ConstantRng(6))


class _BaseGeneratorTestCase(unittest.TestCase):
    def get_game_state(self, **kwargs):
        random.seed(0)
        settings = {"START_WIDTH": 3, "START_HEIGHT": 4, "OBSTACLE_RATIO": 1.0}
        settings.update(kwargs)
        return self.GENERATOR_CLASS(settings).get_game_state(DummyAvatarManager())

    def get_map(self, **kwargs):
        return self.get_game_state(**kwargs).world_map


class TestMainGenerator(_BaseGeneratorTestCase):
    GENERATOR_CLASS = map_generator.Main

    def test_map_dimensions(self):
        m = self.get_map()
        grid = list(m.all_cells())
        self.assertEqual(len(set(grid)), len(grid), "Repeats in list")
        for c in grid:
            self.assertLessEqual(c.location.x, 1)
            self.assertLessEqual(c.location.y, 2)
            self.assertGreaterEqual(c.location.x, -1)
            self.assertGreaterEqual(c.location.y, -1)

    def test_obstacle_ratio(self):
        m = self.get_map(OBSTACLE_RATIO=0)
        obstacle_cells = [cell for cell in m.all_cells() if not cell.habitable]
        self.assertEqual(len(obstacle_cells), 0)

    def test_map_contains_some_non_habitable_cell(self):
        m = self.get_map()
        obstacle_cells = [cell for cell in m.all_cells() if not cell.habitable]
        self.assertGreaterEqual(len(obstacle_cells), 1)

    def test_map_contains_some_habitable_cell_on_border(self):
        m = self.get_map(START_WIDTH=4)
        edge_coordinates = [
            (-1, 2),
            (0, 2),
            (1, 2),
            (2, 2),
            (-1, 1),
            (2, 1),
            (-1, 0),
            (2, 0),
            (-1, -1),
            (0, -1),
            (1, -1),
            (2, -1),
        ]
        edge_cells = (m.get_cell_by_coords(x, y) for (x, y) in edge_coordinates)
        habitable_edge_cells = [cell for cell in edge_cells if cell.habitable]

        self.assertGreaterEqual(len(habitable_edge_cells), 1)

    def test_shortest_path(self):
        m = self.get_map(START_WIDTH=4)


class TestLevel1Generator(_BaseGeneratorTestCase):
    GENERATOR_CLASS = map_generator.Level1

    def test_width_5(self):
        self.assertEqual(self.get_map().num_cols, 5)

    def test_height_1(self):
        self.assertEqual(self.get_map().num_rows, 1)

    def test_static_spawn(self):
        game_state = self.get_game_state()
        sim_runner = SequentialSimulationRunner(game_state, None)
        for i in range(5):
            sim_runner.add_avatar(i)
            self.assertEqual(
                game_state.avatar_manager.avatars_by_id[i].location, Location(-2, 0)
            )



import asyncio
import concurrent.futures
import logging
import threading
import time

from prometheus_client import Histogram

from metrics import GAME_TURN_TIME
from simulation.avatar.avatar_manager import AvatarManager
from simulation.django_communicator import DjangoCommunicator
from simulation.simulation_runner import ConcurrentSimulationRunner
from simulation.worker_manager import WorkerManager

LOGGER = logging.getLogger(__name__)

TURN_TIME = 2


class GameRunner:
    def __init__(
        self,
        game_state_generator,
        communicator: DjangoCommunicator,
        port,
        worker_manager_class=WorkerManager,
    ):
        super(GameRunner, self).__init__()

        self.worker_manager = worker_manager_class(port=port)
        self.game_state = game_state_generator(AvatarManager())
        self.communicator = communicator
        self.simulation_runner = ConcurrentSimulationRunner(
            communicator=self.communicator, game_state=self.game_state
        )
        self._end_turn_callback = lambda: None

    def set_end_turn_callback(self, callback_method):
        self._end_turn_callback = callback_method

    def get_users_to_add(self, game_metadata):
        def player_is_new(_player):
            return _player["id"] not in self.worker_manager.player_id_to_worker.keys()

        return [
            player["id"] for player in game_metadata["users"] if player_is_new(player)
        ]

    def get_users_to_delete(self, game_metadata):
        def player_in_worker_manager_but_not_metadata(pid):
            return pid not in [player["id"] for player in game_metadata["users"]]

        return [
            player_id
            for player_id in self.worker_manager.player_id_to_worker.keys()
            if player_in_worker_manager_but_not_metadata(player_id)
        ]

    def update_main_user(self, game_metadata):
        self.game_state.main_avatar_id = game_metadata["main_avatar"]

    async def update_workers(self):
        game_metadata = self.communicator.get_game_metadata()

        users_to_add = self.get_users_to_add(game_metadata)
        users_to_delete = self.get_users_to_delete(game_metadata)

        await self.worker_manager.add_workers(users_to_add)
        await self.worker_manager.delete_workers(users_to_delete)
        self.simulation_runner.add_avatars(users_to_add)
        self.simulation_runner.delete_avatars(users_to_delete)
        await self.worker_manager.update_worker_codes(game_metadata["users"])

        self.update_main_user(game_metadata)
        await self.worker_manager.fetch_all_worker_data(
            self.game_state.get_serialized_game_states_for_workers()
        )

    async def update_simulation(self, player_id_to_serialized_actions):
        await self.simulation_runner.run_single_turn(player_id_to_serialized_actions)
        await self._end_turn_callback()

    async def update(self):
        with GAME_TURN_TIME():
            await self.update_workers()
            await self.update_simulation(
                self.worker_manager.get_player_id_to_serialized_actions()
            )
            self.worker_manager.clear_logs()
            self.game_state.turn_count += 1

    async def run(self):
        while True:
            LOGGER.info(f"Starting turn {self.game_state.turn_count}")
            turn = asyncio.ensure_future(self.update())
            await asyncio.sleep(TURN_TIME)
            await turn



from django.apps import apps
from django.db import connection
from django.db.migrations.executor import MigrationExecutor
from django.test import TestCase


class MigrationTestCase(TestCase):
    """A Test case for testing migrations."""

    # These must be defined by subclasses.
    start_migration = None
    dest_migration = None

    django_application = None

    @property
    def app_name(self):
        return apps.get_containing_app_config(type(self).__module__).name

    def setUp(self):
        executor = MigrationExecutor(connection)
        # Migrate to start_migration (the migration before the one you want to test)
        executor.migrate([(self.app_name, self.start_migration)])

        # Rebuild graph. Done between invocations of migrate()
        executor.loader.build_graph()

        # Run the migration you want to test
        executor.migrate([(self.app_name, self.dest_migration)])

        # This application can now be used to get the latest models for testing
        self.django_application = executor.loader.project_state(
            [(self.app_name, self.dest_migration)]
        ).apps



import asyncio
import logging
import os
from asyncio import Future
from concurrent import futures
from threading import Thread
from typing import Tuple

from eventlet.semaphore import Semaphore

from simulation.workers import WORKER

LOGGER = logging.getLogger(__name__)
WORKER_TIMEOUT_TIME_SECONDS = 1


class WorkerManager(object):
    """
    Methods of this class must be thread safe unless explicitly stated.
    """

    def __init__(self, port=5000):
        self.player_id_to_worker = {}
        self.port = port
        self.worker_class = WORKER[os.environ.get("WORKER", "local")]

    def get_code(self, player_id):
        return self.player_id_to_worker[player_id].code

    async def fetch_all_worker_data(self, player_id_to_game_state):
        """
        Creates a thread for each worker to send a request for their data. After
        a set duration these threads will close, giving a consistent turn time.
        """
        worker_game_states = [
            (self.player_id_to_worker[player_id], player_id_to_game_state[player_id])
            for player_id in player_id_to_game_state.keys()
        ]

        requests = [
            worker.fetch_data(game_state) for worker, game_state in worker_game_states
        ]

        return asyncio.wait_for(asyncio.gather(*requests), WORKER_TIMEOUT_TIME_SECONDS)

    def get_player_id_to_serialized_actions(self):
        return {
            player_id: self.player_id_to_worker[player_id].serialized_action
            for player_id in self.player_id_to_worker
        }

    def clear_logs(self):
        for worker in self.player_id_to_worker.values():
            worker.log = None

    def update_code(self, player):
        self.player_id_to_worker[player["id"]].code = player["code"]

    def add_new_worker(self, player_id):
        self.player_id_to_worker[player_id] = self.worker_class(player_id, self.port)

    async def _parallel_map(self, func, iterable_args):
        loop = asyncio.get_event_loop()
        with futures.ThreadPoolExecutor() as executor:
            workers: Tuple[Future] = (
                loop.run_in_executor(executor, func, args) for args in iterable_args
            )
            await asyncio.gather(*workers)

    async def add_workers(self, users_to_add):
        await self._parallel_map(self.add_new_worker, users_to_add)

    async def delete_workers(self, players_to_delete):
        await self._parallel_map(self.delete_worker, players_to_delete)

    def delete_worker(self, player_id):
        if player_id in self.player_id_to_worker:
            worker = self.player_id_to_worker[player_id]
            del self.player_id_to_worker[player_id]
            worker.remove_worker()

    async def update_worker_codes(self, players):
        await self._parallel_map(self.update_code, players)



import os

from urllib.parse import urlparse

from simulation.workers.local_worker import LocalWorker
from simulation.worker_manager import WorkerManager

import pytest


def test_local_worker_ports_do_not_conflict(mocker):
    mocker.patch("docker.from_env")
    os.environ["GAME_ID"] = "1"
    LocalWorker._init_port_counter()
    worker_manager1 = WorkerManager()
    worker_manager1.add_new_worker(1)
    local_worker1 = worker_manager1.player_id_to_worker[1]
    url1 = urlparse(local_worker1.url)

    os.environ["GAME_ID"] = "2"
    LocalWorker._init_port_counter()
    worker_manager2 = WorkerManager()
    worker_manager2.add_new_worker(1)
    local_worker2 = worker_manager2.player_id_to_worker[1]
    url2 = urlparse(local_worker2.url)

    assert url1.port == 11989
    assert url2.port == 21989


def test_local_worker_in_the_same_game_do_not_have_port_conflicts(mocker):
    mocker.patch("docker.from_env")
    os.environ["GAME_ID"] = "1"
    LocalWorker._init_port_counter()
    worker_manager = WorkerManager()
    worker_manager.add_new_worker(1)
    worker_manager.add_new_worker(2)
    local_worker1 = worker_manager.player_id_to_worker[1]
    local_worker2 = worker_manager.player_id_to_worker[2]

    url1 = urlparse(local_worker1.url)
    url2 = urlparse(local_worker2.url)

    assert url1.port == 11989
    assert url2.port == 11990



# -*- coding: utf-8 -*-
# Code for Life
#
# Copyright (C) 2015, Ocado Innovation Limited
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ADDITIONAL TERMS – Section 7 GNU General Public Licence
#
# This licence does not grant any right, title or interest in any “Ocado” logos,
# trade names or the trademark “Ocado” or any other trademarks or domain names
# owned by Ocado Innovation Limited or the Ocado group of companies or any other
# distinctive brand features of “Ocado” as may be secured from time to time. You
# must not distribute any modification of this program using the trademark
# “Ocado” or claim any affiliation or association with Ocado or its employees.
#
# You are not authorised to use the name Ocado (or any of its trade names) or
# the names of any author or contributor in advertising or for publicity purposes
# pertaining to the distribution of this program, without the prior written
# authorisation of Ocado.
#
# Any propagation, distribution or conveyance of this program must include this
# copyright notice and these terms. You must not misrepresent the origins of this
# program; modified versions of the program must be marked as such and not
# identified as the original program.
"""aimmo autoconfig"""

DEFAULT_SETTINGS = {"AUTOCONFIG_INDEX_VIEW": "aimmo/home", "STATIC_URL": "/static/"}

SETTINGS = {
    "INSTALLED_APPS": [
        "django.contrib.auth",
        "django.contrib.messages",
        "django.contrib.staticfiles",
        "django_js_reverse",
        "rest_framework",
    ],
    "TEMPLATES": [
        {
            "BACKEND": "django.template.backends.django.DjangoTemplates",
            "APP_DIRS": True,
            "OPTIONS": {
                "context_processors": [
                    "django.template.context_processors.debug",
                    "django.template.context_processors.request",
                    "django.contrib.auth.context_processors.auth",
                    "django.contrib.messages.context_processors.messages",
                ]
            },
        }
    ],
    "USE_TZ": True,
}



from unittest import TestCase

from simulation.cell import Cell

from .test_world_map import serializer


class TestCell(TestCase):
    def test_equal(self):
        cell1 = Cell(1)
        cell2 = Cell(1)
        self.assertEqual(cell1, cell2)

    def test_not_equal(self):
        cell1 = Cell(1)
        cell2 = Cell(2)
        self.assertNotEqual(cell1, cell2)

    def _create_full_cell(self):
        cell = Cell(serializer("location"), False, False)
        cell.avatar = serializer("avatar")
        cell.interactable = serializer("interactable")
        self.expected = {
            "avatar": "avatar",
            "habitable": False,
            "location": "location",
            "interactable": "interactable",
            "partially_fogged": False,
        }
        return cell

    def test_serialize(self):
        cell = self._create_full_cell()
        self.assertEqual(cell.serialize(), self.expected)

    def test_serialize_no_avatar(self):
        cell = self._create_full_cell()
        cell.avatar = None
        self.expected["avatar"] = None
        self.assertEqual(cell.serialize(), self.expected)

    def test_serialize_no_pickup(self):
        cell = self._create_full_cell()
        cell.interactable = None
        self.expected["interactable"] = None
        self.assertEqual(cell.serialize(), self.expected)



class MockCommunicator(object):
    def __init__(self):
        self.data = {
            "main_avatar": 1,
            "users": [
                {
                    "id": 1,
                    "code": "class Avatar(object):\n"
                    "    def next_turn(self, world_view, events):\n"
                    "        from simulation.action import WaitAction\n"
                    "        return WaitAction()\n",
                },
                {
                    "id": 2,
                    "code": "class Avatar(object):\n"
                    "    def next_turn(self, world_view, events):\n"
                    "        from simulation.action import WaitAction\n"
                    "        return WaitAction()\n",
                },
            ],
        }

    def get_game_metadata(self):
        return self.data

    def mark_game_complete(self):
        return {}

    def patch_game(self, data):
        return data

    def change_code(self, avatar_id, new_code):
        users = self.data["users"]
        for i in range(len(users)):
            if users[i]["id"] == avatar_id:
                users[i]["code"] = new_code



import json

from django.http import JsonResponse
from rest_framework import serializers


class GameSerializer(serializers.Serializer):
    name = serializers.CharField(max_length=100)
    settings = serializers.SerializerMethodField("get_settings_as_dict")
    status = serializers.CharField(max_length=1)

    def get_settings_as_dict(self, game):
        return json.dumps(game.settings_as_dict(), sort_keys=True)

    def update(self, instance, validated_data):
        instance.name = validated_data.get("name", instance.name)
        instance.status = validated_data.get("status", instance.status)
        instance.save()
        return instance



from __future__ import absolute_import

import logging
import os
import sys
import time

import django
from django.conf import settings

from . import docker_scripts
from .shell_api import log, run_command, run_command_async

sys.path.append("/home/travis/build/ocadotechnology/aimmo")

try:
    if os.environ["CI"] == "true":
        ROOT_DIR_LOCATION = os.environ["TRAVIS_BUILD_DIR"]
    else:
        ROOT_DIR_LOCATION = os.path.abspath(
            os.path.dirname((os.path.dirname(__file__)))
        )
except KeyError:
    ROOT_DIR_LOCATION = os.path.abspath(os.path.dirname((os.path.dirname(__file__))))

_MANAGE_PY = os.path.join(ROOT_DIR_LOCATION, "example_project", "manage.py")
_SERVICE_PY = os.path.join(ROOT_DIR_LOCATION, "aimmo-game-creator", "service.py")
_FRONTEND_BUNDLER_JS = os.path.join(
    ROOT_DIR_LOCATION, "game_frontend", "djangoBundler.js"
)

PROCESSES = []


def create_superuser_if_missing(username, password):
    from django.contrib.auth.models import User

    try:
        User.objects.get_by_natural_key(username)
    except User.DoesNotExist:
        log("Creating superuser %s with password %s" % (username, password))
        User.objects.create_superuser(
            username=username, email="admin@admin.com", password=password
        )


def run(
    use_minikube,
    server_wait=True,
    capture_output=False,
    test_env=False,
    build_target=None,
):
    logging.basicConfig()
    if test_env:
        os.environ.setdefault("DJANGO_SETTINGS_MODULE", "test_settings")
    else:
        sys.path.append(os.path.join(ROOT_DIR_LOCATION, "example_project"))
        os.environ.setdefault("DJANGO_SETTINGS_MODULE", "example_project.settings")

    django.setup()
    run_command(
        ["pip", "install", "-e", ROOT_DIR_LOCATION], capture_output=capture_output
    )

    if not test_env:
        run_command(
            ["python", _MANAGE_PY, "migrate", "--noinput"],
            capture_output=capture_output,
        )
        run_command(
            ["python", _MANAGE_PY, "collectstatic", "--noinput"],
            capture_output=capture_output,
        )

    create_superuser_if_missing(username="admin", password="admin")

    server_args = []
    if use_minikube:
        parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        sys.path.append(os.path.join(parent_dir, "aimmo_runner"))

        os.chdir(ROOT_DIR_LOCATION)
        run_command(
            [
                "pip",
                "install",
                "-r",
                os.path.join(ROOT_DIR_LOCATION, "minikube_requirements.txt"),
            ],
            capture_output=capture_output,
        )

        # Import minikube here, so we can install the dependencies first
        from aimmo_runner import minikube

        minikube.start(build_target=build_target)

        server_args.append("0.0.0.0:8000")
        os.environ["AIMMO_MODE"] = "minikube"
    else:
        time.sleep(2)
        os.environ["AIMMO_MODE"] = "threads"
        docker_scripts.delete_containers()
        if build_target == "tester":
            run_command(["python", "all_tests.py"])
        else:
            docker_scripts.build_docker_images(build_target=build_target)
            docker_scripts.start_game_creator()

    os.environ["NODE_ENV"] = "development" if settings.DEBUG else "production"
    server = run_command_async(
        ["python", _MANAGE_PY, "runserver"] + server_args, capture_output=capture_output
    )
    frontend_bundler = run_command_async(
        ["node", _FRONTEND_BUNDLER_JS], capture_output=capture_output
    )
    PROCESSES.append(server)
    PROCESSES.append(frontend_bundler)

    if server_wait is True:
        try:
            game.wait()
        except NameError:
            pass

        server.wait()

    return PROCESSES



from simulation.workers.worker import Worker


class ConcreteWorker(Worker):
    def __init__(self, *args, **kwargs):
        super(ConcreteWorker, self).__init__(*args, **kwargs)
        self.code = """
def next_turn(world_state, avatar_state):
    return MoveAction(direction.NORTH)
"""

    def _create_worker(self):
        return "http://test"

    def remove_worker(self):
        pass



from enum import Enum
from threading import RLock

from simulation.interactables import serialize_interactables


class GameState:
    """
    Encapsulates the entire game state, including avatars, their code, and the world.
    """

    def __init__(
        self, world_map, avatar_manager, completion_check_callback=lambda: None
    ):
        self.world_map = world_map
        self.avatar_manager = avatar_manager
        self.main_avatar_id = None
        self.turn_count = 0
        self._lock = RLock()

    def get_main_avatar(self):
        with self._lock:
            return self.avatar_manager.avatars_by_id[self.main_avatar_id]

    def serialize(self):
        return {
            "era": "less_flat",
            "southWestCorner": self.world_map.get_serialized_south_west_corner(),
            "northEastCorner": self.world_map.get_serialized_north_east_corner(),
            "players": self.avatar_manager.serialize_players(),
            "interactables": serialize_interactables(self.world_map),
            "obstacles": self.world_map.serialize_obstacles(),
        }

    def serialize_for_worker(self, avatar_wrapper):
        with self._lock:
            return {
                "avatar_state": avatar_wrapper.serialize(),
                "world_map": {
                    "cells": [cell.serialize() for cell in self.world_map.all_cells()]
                },
            }

    def get_serialized_game_states_for_workers(self):
        with self._lock:
            return {
                player_id: self.serialize_for_worker(avatar_wrapper)
                for player_id, avatar_wrapper in self.avatar_manager.avatars_by_id.items()
            }



import asyncio
import random

import hypothesis.strategies as st
from hypothesis import given
from simulation.location import Location
from simulation.interactables.pickups import HealthPickup
from simulation.interactables.effects import (
    AVATAR_HEALTH_MAX,
    HEALTH_RESTORE_DEFAULT,
    HEALTH_RESTORE_MAX,
)

from .mock_world import MockWorld


class TestHealthPickupAndEffects:
    def setup_method(self):
        """
        Mock a game for each test individually. MockWorld() will set up a:
        avatar manager, game state, turn manager and a map generator.
        """
        self.game = MockWorld()
        self.game.simulation_runner.add_avatar(1, Location(0, 0))
        self.cell = self.game.game_state.world_map.get_cell(Location(1, 0))
        self.initial_health = self.game.avatar_manager.get_avatar(1).health

    async def test_health_pickups_and_effects_apply_default(self, loop):
        """
        HealthPickups without any parameter provided.
        """
        self.cell.interactable = HealthPickup(self.cell)
        assert self.cell.interactable.serialize() == {
            "type": "health",
            "location": {"x": self.cell.location.x, "y": self.cell.location.y},
        }

        await self.game.simulation_runner.run_single_turn(
            self.game.avatar_manager.get_player_id_to_serialized_action()
        )

        assert self.cell.avatar == self.game.avatar_manager.get_avatar(1)
        assert self.cell.avatar.health == self.initial_health + HEALTH_RESTORE_DEFAULT

    async def test_health_effect_is_capped_at_HEALTH_RESTORE_MAX(self, loop):
        """
        Make sure health cannot go above the maximum cap. Avatar health begins at 5hp,
        so every pickup in the 96-HEALTH_RESTORE_MAX range would cause it to go above
        HEALTH_RESTORE_MAX, until the initial health changes.
        """
        avatar = self.game.avatar_manager.get_avatar(1)
        avatar.health = 97
        self.cell.interactable = HealthPickup(self.cell)

        await self.game.simulation_runner.run_single_turn(
            self.game.avatar_manager.get_player_id_to_serialized_action()
        )

        assert self.cell.avatar == self.game.avatar_manager.get_avatar(1)
        assert self.cell.avatar.health == AVATAR_HEALTH_MAX



# -*- coding: utf-8 -*-
import os

from setuptools import find_packages, setup

withcoverage = os.environ.get("WITH_COVERAGE")

setup(
    name="aimmo-game",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "aiohttp",
        "aiohttp-cors",
        "aiohttp-wsgi",
        "eventlet",
        "python-socketio==4.0.0",
        "requests",
        "six",
        "kubernetes",
        "prometheus_client",
        "docker",
    ],
    setup_requires=["pytest-runner"],
    tests_require=[
        "socketIO-client",
        "pytest>=5.2, <6",
        "pytest-asyncio==0.10.0",
        "pytest-mock",
        "asynctest",
        "httmock",
        "mock",
        "hypothesis",
        "pytest-aiohttp",
        "aioresponses",
    ],
    test_suite="tests",
    zip_safe=False,
)



import asyncio
import math

from hypothesis import assume, given
from hypothesis import strategies as st
from simulation.location import Location
from simulation.interactables.pickups import DamageBoostPickup
from simulation.interactables.effects import DAMAGE_BOOST_DEFAULT

from .mock_world import MockWorld


class TestDamagePickupsAndEffects:
    def setup_method(self):
        """
        Mock a game for each test individually. MockWorld() will set up a:
        avatar manager, game state, turn manager and a map generator.
        """
        self.game = MockWorld()
        self.game.simulation_runner.add_avatar(1, Location(0, 0))
        _avatar_spawn_cell = self.game.game_state.world_map.get_cell(Location(0, 0))
        self.initial_attack_strength = _avatar_spawn_cell.avatar.attack_strength
        self.cell = self.game.game_state.world_map.get_cell(Location(1, 0))

    async def test_damage_boost_pickup_can_be_picked_up_default(self, loop):
        pickup_created = DamageBoostPickup(self.cell)
        self.cell.interactable = pickup_created
        assert self.cell.interactable.serialize() == {
            "type": "damage_boost",
            "location": {"x": self.cell.location.x, "y": self.cell.location.y},
        }
        await self.game.simulation_runner.run_single_turn(
            self.game.avatar_manager.get_player_id_to_serialized_action()
        )

        assert self.cell.avatar == self.game.avatar_manager.get_avatar(1)
        assert len(self.game.avatar_manager.get_avatar(1).effects) == 1
        damage_boost_effect = self.game.avatar_manager.get_avatar(1).effects.pop()
        assert isinstance(damage_boost_effect, pickup_created.effects[0])

    async def test_damage_boost_increases_attack_strength_with_default_integer(
        self, loop
    ):
        """
        Damage boost with no value parameter provided (ie. default).
        """
        pickup_created = DamageBoostPickup(self.cell)
        self.cell.interactable = pickup_created

        await self.game.simulation_runner.run_single_turn(
            self.game.avatar_manager.get_player_id_to_serialized_action()
        )

        assert (
            self.cell.avatar.attack_strength
            == self.initial_attack_strength + DAMAGE_BOOST_DEFAULT
        )



"""
This module contains all currently availible effects.

Effects are applied to all targets specified by a pickup
When creating effects they should follow the same basic format:

```python

class NewEffect(_Effect):
    def __init__(self, *args):
        super(NewEffect, self).__init__(duration=X,*args)
        # what the effect does goes here

    def remove(self):
        super(NewEffect, self).remove()
        # if the effect is temporary, it should be undone here
```
"""
import math
from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from simulation.avatar.avatar_wrapper import AvatarWrapper

DEFAULT_EFFECT_TIME = 10

INVULNERABILITY_RESISTANCE = 1000
DAMAGE_BOOST_DEFAULT = 5

HEALTH_RESTORE_DEFAULT = 3
HEALTH_RESTORE_MAX = 100
AVATAR_HEALTH_MAX = 100

SCORE_INCREASE_DEFAULT = 1


class _Effect(object):
    """
    Base effect class, does nothing on its own.
    """

    def __init__(self, recipient: "AvatarWrapper", duration=DEFAULT_EFFECT_TIME):
        self._recipient = recipient
        self.is_expired = False
        self._time_remaining = duration
        try:
            self._recipient.effects.add(self)
        except KeyError as e:
            raise KeyError("The target object does support effects.")

    def on_turn(self):
        self._time_remaining -= 1
        if self._time_remaining <= 0:
            self.is_expired = True

    def remove(self):
        try:
            self._recipient.effects.remove(self)
        except KeyError as e:
            raise KeyError(
                "The target object does not exist! Cannot remove the effect."
            )


class InvulnerabilityEffect(_Effect):
    def __init__(self, *args):
        super(InvulnerabilityEffect, self).__init__(*args)
        self._recipient.resistance += INVULNERABILITY_RESISTANCE

    def remove(self):
        super(InvulnerabilityEffect, self).remove()
        self._recipient.resistance -= INVULNERABILITY_RESISTANCE

    def __repr__(self):
        return f"InvulnerabilityEffect(value={INVULNERABILITY_RESISTANCE})"


class DamageBoostEffect(_Effect):
    def __init__(self, *args):
        self._damage_boost = int(round(DAMAGE_BOOST_DEFAULT))
        super(DamageBoostEffect, self).__init__(*args)
        self._recipient.attack_strength += self._damage_boost

    def remove(self):
        super(DamageBoostEffect, self).remove()
        self._recipient.attack_strength -= self._damage_boost

    def __repr__(self):
        return f"DamageBoostEffect(value={self._damage_boost})"


class HealthEffect(_Effect):
    def __init__(self, *args):
        super(HealthEffect, self).__init__(duration=1, *args)
        self.health_restored = HEALTH_RESTORE_DEFAULT
        self.give_health(self._recipient)

    def remove(self):
        super(HealthEffect, self).remove()

    def give_health(self, avatar):
        avatar.health += self.health_restored

        # Make sure the health is capped at 100.
        if avatar.health > AVATAR_HEALTH_MAX:
            avatar.health = AVATAR_HEALTH_MAX

    def __repr__(self):
        return f"HealthEffect(value={self.health_restored})"


class ScoreEffect(_Effect):
    def __init__(self, *args):
        super(ScoreEffect, self).__init__(duration=1, *args)
        self.score_given = SCORE_INCREASE_DEFAULT
        self._recipient.score += self.score_given

    def remove(self):
        super(ScoreEffect, self).remove()

    def __repr__(self):
        return f"ScoreEffect(value={self.score_given})"



from simulation.worker_manager import WorkerManager
from .concrete_worker import ConcreteWorker


class MockWorkerManager(WorkerManager):
    def __init__(self, *args, **kwargs):
        self.final_workers = set()
        self.updated_workers = []

        super(MockWorkerManager, self).__init__(*args, **kwargs)
        self.worker_class = ConcreteWorker

    def add_new_worker(self, player_id):
        self.final_workers.add(player_id)
        super(MockWorkerManager, self).add_new_worker(player_id)

    def delete_worker(self, player_id):
        try:
            self.final_workers.remove(player_id)
        except KeyError:
            pass
        super(MockWorkerManager, self).delete_worker(player_id)

    def update_code(self, user):
        self.updated_workers.append(user["id"])
        super(MockWorkerManager, self).update_code(user)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models, transaction, IntegrityError

import aimmo.models


def migrate_data_forward(apps, schema_editor):
    Player = apps.get_model("aimmo", "Player")
    Avatar = apps.get_model("aimmo", "Avatar")
    Game = apps.get_model("aimmo", "Game")

    if Player.objects.count() == 0:
        return
    main_game = Game(pk=1, name="main")
    main_game.save()

    avatars = [
        Avatar(game=main_game, owner=player.user, code=player.code)
        for player in Player.objects.all()
    ]
    Avatar.objects.bulk_create(avatars)


def migrate_data_backward(apps, schema_editor):
    Player = apps.get_model("aimmo", "Player")
    Avatar = apps.get_model("aimmo", "Avatar")

    if Avatar.objects.count() == 0:
        return
    for avatar in Avatar.objects.all():
        player = Player(user=avatar.owner, code=avatar.code)
        try:
            with transaction.atomic():
                player.save()
        except IntegrityError:
            # Stuff doesn't map as can have more than one Avatar but only one Player
            pass


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("aimmo", "0002_auto_20160601_1914"),
    ]

    operations = [
        migrations.CreateModel(
            name="Avatar",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("code", models.TextField()),
                (
                    "auth_token",
                    models.CharField(
                        default=aimmo.models.generate_auth_token, max_length=24
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Game",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("name", models.CharField(max_length=100)),
                (
                    "auth_token",
                    models.CharField(
                        default=aimmo.models.generate_auth_token, max_length=24
                    ),
                ),
                ("public", models.BooleanField(default=True)),
                (
                    "can_play",
                    models.ManyToManyField(
                        related_name="playable_games", to=settings.AUTH_USER_MODEL
                    ),
                ),
                (
                    "owner",
                    models.ForeignKey(
                        related_name="owned_games",
                        to=settings.AUTH_USER_MODEL,
                        null=True,
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="avatar", name="game", field=models.ForeignKey(to="aimmo.Game")
        ),
        migrations.AddField(
            model_name="avatar",
            name="owner",
            field=models.ForeignKey(to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterUniqueTogether(
            name="avatar", unique_together=set([("owner", "game")])
        ),
        migrations.RunPython(migrate_data_forward, migrate_data_backward),
        migrations.RemoveField(model_name="player", name="user"),
        migrations.DeleteModel(name="Player"),
    ]



from __future__ import absolute_import
from .base_test_migration import MigrationTestCase


class TestMigrationPreviewUsers(MigrationTestCase):

    start_migration = "0008_default_public_field"
    dest_migration = "0009_add_game_status"

    def test_game_status_field_added(self):
        model = self.django_application.get_model(self.app_name, "Game")

        self.assertEquals(
            model._meta.get_field("status").get_internal_type(), "CharField"
        )



from __future__ import absolute_import

import asyncio
from string import ascii_uppercase
from unittest.mock import patch

from simulation.avatar.avatar_appearance import AvatarAppearance
from simulation.game_state import GameState
from simulation.interactables.pickups import DamageBoostPickup
from simulation.interactables.score_location import ScoreLocation
from simulation.location import Location
from simulation.simulation_runner import ConcurrentSimulationRunner
from simulation.world_map import WorldMap

from .dummy_avatar import (
    DeadDummy,
    DummyAvatar,
    DummyAvatarManager,
    MoveEastDummy,
    MoveNorthDummy,
    MoveSouthDummy,
    MoveWestDummy,
    WaitDummy,
)
from .maps import InfiniteMap, MockCell, MockPickup
from .mock_communicator import MockCommunicator

ORIGIN = Location(0, 0)

RIGHT_OF_ORIGIN = Location(1, 0)
FIVE_RIGHT_OF_ORIGIN = Location(5, 0)

ABOVE_ORIGIN = Location(0, 1)
FIVE_RIGHT_OF_ORIGIN_AND_ONE_ABOVE = Location(5, 1)

SETTINGS = {
    "TARGET_NUM_CELLS_PER_AVATAR": 0,
    "TARGET_NUM_PICKUPS_PER_AVATAR": 0,
    "TARGET_NUM_SCORE_LOCATIONS_PER_AVATAR": 0,
    "SCORE_DESPAWN_CHANCE": 0,
    "PICKUP_SPAWN_CHANCE": 0,
}


class MockGameState(GameState):
    def get_state_for(self, avatar):
        return self


class TestSimulationRunner:
    """
        Key:
            > : Avatar moving eastward
            < : Avatar moving westward
            x : Avatar waiting / blocked
            o : Avatar successfully moved
            ! : Dead avatar (that should be waiting)
    """

    def _generate_grid(self, columns=2, rows=2):
        alphabet = iter(ascii_uppercase)
        grid = {
            Location(x, y): MockCell(Location(x, y), name=next(alphabet))
            for x in range(columns)
            for y in range(rows)
        }
        return grid

    def assertGridSize(self, world_map, expected_columns, expected_rows=None):
        if expected_rows is None:
            expected_rows = expected_columns
        assert world_map.num_rows == expected_rows
        assert world_map.num_cols == expected_columns
        assert world_map.num_cells == expected_rows * expected_columns
        assert len(list(world_map.all_cells())) == expected_rows * expected_columns

    def construct_default_avatar_appearance(self):
        return AvatarAppearance("#000", "#ddd", "#777", "#fff")

    def construct_simulation_runner(self, avatars, locations):
        self.avatar_manager = DummyAvatarManager(avatars)
        self.avatar_manager.avatars_by_id = dict(enumerate(avatars))
        self.game_state = MockGameState(InfiniteMap(), self.avatar_manager)
        self.simulation_runner = ConcurrentSimulationRunner(
            game_state=self.game_state, communicator=MockCommunicator()
        )
        for index, location in enumerate(locations):
            self.simulation_runner.add_avatar(index, location)

    def assert_at(self, avatar, location):
        assert avatar.location == location
        cell = self.game_state.world_map.get_cell(location)
        assert cell.avatar == avatar

    def get_avatar(self, player_id):
        return self.avatar_manager.get_avatar(player_id)

    async def run_turn(self):
        await self.simulation_runner.run_turn(self.avatar_manager.avatars_by_id)

    def test_add_avatar(self):
        self.construct_simulation_runner([], [])

        self.simulation_runner.add_avatar(7)
        assert 7 in self.game_state.avatar_manager.avatars_by_id

    def test_remove_avatar(self):
        self.construct_simulation_runner(
            [DummyAvatar, DummyAvatar], [Location(0, 0), Location(1, 1)]
        )

        avatar = self.get_avatar(1)
        avatar.marked = True

        self.simulation_runner.remove_avatar(0)

        assert not 0 in self.avatar_manager.avatars_by_id
        assert self.game_state.world_map.get_cell(Location(0, 0)).avatar == None

        assert self.avatar_manager.avatars_by_id[1].marked
        assert self.game_state.world_map.get_cell(Location(1, 1)).avatar.marked

    def test_remove_non_existent_avatar(self):
        self.construct_simulation_runner([], [])
        self.simulation_runner.remove_avatar(10)

    def test_updates_map_with_correct_num_avatars(self):
        self.construct_simulation_runner([], [])

        self.avatar_manager.add_avatar(1)
        self.simulation_runner.update_environment()
        assert len(self.simulation_runner.game_state.avatar_manager.avatars_by_id) == 1

        self.avatar_manager.add_avatar(2)
        self.avatar_manager.add_avatar(3)
        self.simulation_runner.update_environment()
        assert len(self.simulation_runner.game_state.avatar_manager.avatars_by_id) == 3

    def test_grid_expand(self):
        self.construct_simulation_runner([], [])
        settings = SETTINGS.copy()
        settings["TARGET_NUM_CELLS_PER_AVATAR"] = 5
        self.simulation_runner.game_state.world_map = WorldMap(
            self._generate_grid(), settings
        )
        self.simulation_runner.update(1, self.simulation_runner.game_state)
        print(self.simulation_runner.game_state.world_map)
        assert self.simulation_runner.game_state.world_map.is_on_map(Location(-1, -1))
        assert self.simulation_runner.game_state.world_map.is_on_map(Location(-1, 2))
        assert self.simulation_runner.game_state.world_map.is_on_map(Location(2, 2))
        assert self.simulation_runner.game_state.world_map.is_on_map(Location(2, -1))
        self.assertGridSize(self.simulation_runner.game_state.world_map, 4)

        self.simulation_runner.update(4, self.simulation_runner.game_state)
        self.assertGridSize(self.simulation_runner.game_state.world_map, 6)
        assert self.simulation_runner.game_state.world_map.is_on_map(Location(0, 3))
        assert self.simulation_runner.game_state.world_map.is_on_map(Location(3, 0))
        assert self.simulation_runner.game_state.world_map.is_on_map(Location(-2, 0))
        assert self.simulation_runner.game_state.world_map.is_on_map(Location(0, -2))

    def test_grid_doesnt_expand(self):
        self.construct_simulation_runner([], [])
        settings = SETTINGS.copy()
        settings["TARGET_NUM_CELLS_PER_AVATAR"] = 4
        self.simulation_runner.game_state.world_map = WorldMap(
            self._generate_grid(), settings
        )
        self.simulation_runner.update(1, self.simulation_runner.game_state)
        self.assertGridSize(self.simulation_runner.game_state.world_map, 2)

    def test_score_despawn_chance(self):
        self.construct_simulation_runner([], [])
        settings = SETTINGS.copy()
        settings["TARGET_NUM_SCORE_LOCATIONS_PER_AVATAR"] = 0
        grid = self._generate_grid()
        grid[Location(0, 1)].interactable = ScoreLocation(grid[Location(0, 1)])
        self.simulation_runner.game_state.world_map = WorldMap(grid, settings)
        self.simulation_runner.update(1, self.simulation_runner.game_state)
        assert (
            grid[Location(0, 1)]
            in self.simulation_runner.game_state.world_map.score_cells()
        )
        assert len(list(self.simulation_runner.game_state.world_map.score_cells())) == 1

    def test_scores_applied(self):
        self.construct_simulation_runner([], [])
        grid = self._generate_grid()
        avatar = DummyAvatar()
        grid[Location(1, 1)].interactable = ScoreLocation(grid[Location(1, 1)])
        grid[Location(1, 1)].avatar = avatar
        self.simulation_runner.game_state.world_map = WorldMap(grid, SETTINGS)
        self.simulation_runner.update(1, self.simulation_runner.game_state)
        assert avatar.score == 1

    def test_scores_not_added_when_at_target(self):
        self.construct_simulation_runner([], [])
        settings = SETTINGS.copy()
        settings["TARGET_NUM_SCORE_LOCATIONS_PER_AVATAR"] = 1
        grid = self._generate_grid()
        grid[Location(0, 1)].interactable = ScoreLocation(grid[Location(0, 1)])
        self.simulation_runner.game_state.world_map = WorldMap(grid, settings)
        self.simulation_runner.update(1, self.simulation_runner.game_state)
        assert len(list(self.simulation_runner.game_state.world_map.score_cells())) == 1
        assert (
            grid[Location(0, 1)]
            in self.simulation_runner.game_state.world_map.score_cells()
        )

    def test_not_enough_score_space(self):
        self.construct_simulation_runner([], [])
        settings = SETTINGS.copy()
        settings["TARGET_NUM_SCORE_LOCATIONS_PER_AVATAR"] = 1
        grid = self._generate_grid(1, 1)
        grid[Location(0, 0)].avatar = "avatar"
        self.simulation_runner.game_state.world_map = WorldMap(grid, settings)
        self.simulation_runner.update(1, self.simulation_runner.game_state)
        assert len(list(self.simulation_runner.game_state.world_map.score_cells())) == 0

    def test_pickups_applied(self):
        self.construct_simulation_runner([], [])
        grid = self._generate_grid()
        avatar = DummyAvatar()
        pickup = MockPickup(target=avatar)
        grid[Location(1, 1)].interactable = pickup
        grid[Location(1, 1)].avatar = avatar
        self.simulation_runner.game_state.world_map = WorldMap(grid, SETTINGS)
        self.simulation_runner.update(1, self.simulation_runner.game_state)
        assert pickup.applied_to == avatar

    def test_pickup_spawn_chance(self):
        self.construct_simulation_runner([], [])
        settings = SETTINGS.copy()
        settings["TARGET_NUM_PICKUPS_PER_AVATAR"] = 5
        settings["PICKUP_SPAWN_CHANCE"] = 0
        grid = self._generate_grid()
        self.simulation_runner.game_state.world_map = WorldMap(grid, settings)
        self.simulation_runner.update(1, self.simulation_runner.game_state)
        assert (
            len(list(self.simulation_runner.game_state.world_map.interactable_cells()))
            == 0
        )

    @patch("simulation.interactables.pickups.DamageBoostPickup")
    def test_pickups_not_added_when_at_target(self, mockPickup):
        self.construct_simulation_runner([], [])
        settings = SETTINGS.copy()
        settings["TARGET_NUM_PICKUPS_PER_AVATAR"] = 1
        grid = self._generate_grid()
        grid[Location(0, 1)].interactable = mockPickup()
        self.simulation_runner.game_state.world_map = WorldMap(grid, settings)
        self.simulation_runner.update(1, self.simulation_runner.game_state)
        assert (
            len(list(self.simulation_runner.game_state.world_map.interactable_cells()))
            == 1
        )
        assert (
            grid[Location(0, 1)]
            in self.simulation_runner.game_state.world_map.interactable_cells()
        )

    def test_not_enough_pickup_space(self):
        self.construct_simulation_runner([], [])
        settings = SETTINGS.copy()
        settings["TARGET_NUM_PICKUPS_PER_AVATAR"] = 1
        grid = self._generate_grid(1, 1)
        grid[Location(0, 0)].interactable = ScoreLocation(grid[Location(0, 0)])
        self.simulation_runner.game_state.world_map = WorldMap(grid, settings)
        self.simulation_runner.update(1, self.simulation_runner.game_state)
        assert (
            len(list(self.simulation_runner.game_state.world_map.pickup_cells())) == 0
        )

    async def test_run_turn(self, loop):
        """
        Given:  > _
        (1)
        Expect: _ o
        """
        self.construct_simulation_runner([MoveEastDummy], [ORIGIN])
        avatar = self.get_avatar(0)

        self.assert_at(avatar, ORIGIN)
        await self.run_turn()
        self.assert_at(avatar, RIGHT_OF_ORIGIN)

    async def test_run_several_turns(self, loop):
        """
        Given:  > _ _ _ _ _
        (5)
        Expect: _ _ _ _ _ o
        """
        self.construct_simulation_runner([MoveEastDummy], [ORIGIN])
        avatar = self.get_avatar(0)

        assert avatar.location == ORIGIN
        [await self.run_turn() for _ in range(5)]
        assert avatar.location == FIVE_RIGHT_OF_ORIGIN

    async def test_run_several_turns_and_avatars(self, loop):
        """
        Given:  > _ _ _ _ _
                > _ _ _ _ _
        (5)
        Expect: _ _ _ _ _ o
                _ _ _ _ _ o
        """
        self.construct_simulation_runner(
            [MoveEastDummy, MoveEastDummy], [ORIGIN, ABOVE_ORIGIN]
        )
        avatar0 = self.get_avatar(0)
        avatar1 = self.get_avatar(1)

        self.assert_at(avatar0, ORIGIN)
        self.assert_at(avatar1, ABOVE_ORIGIN)
        [await self.run_turn() for _ in range(5)]
        self.assert_at(avatar0, FIVE_RIGHT_OF_ORIGIN)
        self.assert_at(avatar1, FIVE_RIGHT_OF_ORIGIN_AND_ONE_ABOVE)

    async def test_move_chain_succeeds(self, loop):
        """
        Given:  > > > > > _

        Expect: _ o o o o o
        """
        self.construct_simulation_runner(
            [MoveEastDummy for _ in range(5)], [Location(x, 0) for x in range(5)]
        )
        avatars = [self.get_avatar(i) for i in range(5)]

        [self.assert_at(avatars[x], Location(x, 0)) for x in range(5)]
        await self.run_turn()
        [self.assert_at(avatars[x], Location(x + 1, 0)) for x in range(5)]

    async def test_move_chain_fails_occupied(self, loop):
        """
        Given:  > > x _

        Expect: x x x _
        """
        self.construct_simulation_runner(
            [MoveEastDummy, MoveEastDummy, WaitDummy],
            [Location(x, 0) for x in range(3)],
        )
        avatars = [self.get_avatar(i) for i in range(3)]

        [self.assert_at(avatars[x], Location(x, 0)) for x in range(3)]
        await self.run_turn()
        [self.assert_at(avatars[x], Location(x, 0)) for x in range(3)]

    async def test_move_chain_fails_occupied_by_dead_avatar(self, loop):
        """
        Given: > > ! _

        Expect: x x ! _
        """

        self.construct_simulation_runner(
            [MoveEastDummy, MoveEastDummy, DeadDummy],
            [Location(x, 0) for x in range(3)],
        )
        avatars = [self.get_avatar(i) for i in range(3)]

        [self.assert_at(avatars[x], Location(x, 0)) for x in range(3)]
        await self.run_turn()
        [self.assert_at(avatars[x], Location(x, 0)) for x in range(3)]

    async def test_move_fails_collision(self, loop):
        """
        Given: > _ <
        Expect: x _ x
        """

        self.construct_simulation_runner(
            [MoveEastDummy, MoveWestDummy], [Location(0, 0), Location(2, 0)]
        )
        avatars = [self.get_avatar(i) for i in range(2)]

        self.assert_at(avatars[0], Location(0, 0))
        self.assert_at(avatars[1], Location(2, 0))

        await self.run_turn()

        self.assert_at(avatars[0], Location(0, 0))
        self.assert_at(avatars[1], Location(2, 0))

    async def test_move_chain_fails_collision(self, loop):
        """
        Given:  > > > _ <
        (1)
        Expect: x x x _ x
        """
        locations = [Location(0, 0), Location(1, 0), Location(2, 0), Location(4, 0)]
        self.construct_simulation_runner(
            [MoveEastDummy, MoveEastDummy, MoveEastDummy, MoveWestDummy], locations
        )
        avatars = [self.get_avatar(i) for i in range(4)]

        [self.assert_at(avatars[i], locations[i]) for i in range(4)]
        await self.run_turn()
        [self.assert_at(avatars[i], locations[i]) for i in range(4)]

    async def test_move_chain_fails_cycle(self, loop):
        """
        Given:  > v
                ^ <
        (1)
        Expect: x x
                x x
        """
        locations = [Location(0, 1), Location(1, 1), Location(1, 0), Location(0, 0)]
        self.construct_simulation_runner(
            [MoveEastDummy, MoveSouthDummy, MoveWestDummy, MoveNorthDummy], locations
        )
        avatars = [self.get_avatar(i) for i in range(4)]

        [self.assert_at(avatars[i], locations[i]) for i in range(4)]
        await self.run_turn()
        [self.assert_at(avatars[i], locations[i]) for i in range(4)]

    async def test_move_chain_fails_spiral(self, loop):
        """
        Given:  > > v
                  ^ <
        (1)
        Expect: x x x
                  x x
        """
        locations = [
            Location(0, 1),
            Location(1, 1),
            Location(2, 1),
            Location(2, 0),
            Location(1, 0),
        ]
        self.construct_simulation_runner(
            [
                MoveEastDummy,
                MoveEastDummy,
                MoveSouthDummy,
                MoveWestDummy,
                MoveNorthDummy,
            ],
            locations,
        )
        avatars = [self.get_avatar(i) for i in range(5)]

        [self.assert_at(avatars[i], locations[i]) for i in range(5)]
        await self.run_turn()
        [self.assert_at(avatars[i], locations[i]) for i in range(5)]


if __name__ == "__main__":
    unittest.main()



from __future__ import absolute_import

from rest_framework import authentication, permissions

from .app_settings import CanDelete


class CsrfExemptSessionAuthentication(authentication.SessionAuthentication):
    def enforce_csrf(self, request):
        return False


class GameHasToken(permissions.BasePermission):
    """
    Used to verify that an incoming request has permission
    to access a given object from the models.

    This is done on a per object basis. The object must have an `auth_token`
    attribute to be used with this permission class.
    """

    def has_permission(self, request, view):
        return True

    def has_object_permission(self, request, view, obj):
        return self.check_for_token(request, obj)

    def check_for_token(self, request, obj):
        try:
            return (
                obj.auth_token == ""
                or request.META["HTTP_GAME_TOKEN"] == obj.auth_token
            )
        except (KeyError, AttributeError):
            return False


class CanUserPlay(permissions.BasePermission):
    """
    Used to verify that an incoming request is made by a user
    that's authorised to play an AIMMO game
    """

    def has_permission(self, request, view):
        return True

    def has_object_permission(self, request, view, obj):
        return obj.can_user_play(request.user)


class CanDeleteGameOrReadOnly(permissions.BasePermission):
    """
    Used to verify that an incoming request is made by a user
    that's authorised to delete or view an AIMMO game
    """

    def has_permission(self, request, view):
        return True

    def has_object_permission(self, request, view, obj):
        if request.method in permissions.SAFE_METHODS:
            return True
        elif request.method == "PATCH":
            return GameHasToken().has_object_permission(request, view, obj)
        else:
            can_play = CanUserPlay().has_object_permission(request, view, obj)
            return CanDelete().has_permission(request, view) and can_play



from __future__ import absolute_import

import asyncio
import json

from tests.test_simulation.concrete_worker import ConcreteWorker

from httmock import HTTMock
from simulation.avatar import avatar_wrapper
from simulation.location import Location


class MockEffect(object):
    def __init__(self, avatar):
        self.turns = 0
        self.is_expired = False
        self.removed = False
        self.expire = False
        self.avatar = avatar

    def on_turn(self):
        self.is_expired = self.expire
        self.turns += 1

    def remove(self):
        self.removed = True
        self.avatar.effects.remove(self)


class MockAction(object):
    def __init__(self, avatar, **options):
        global actions_created
        self.options = options
        self.avatar = avatar
        actions_created.append(self)


class ActionRequest(object):
    def __init__(self, options=None):
        if options is None:
            options = {}
        self.options = options

    def __call__(self, url, request):
        return json.dumps({"action": {"action_type": "test", "options": self.options}})


def InvalidJSONRequest(url, request):
    return "EXCEPTION"


def NonExistentActionRequest(url, request):
    return json.dumps({"action": {"action_type": "fake", "option": {}}})


avatar_wrapper.ACTIONS = {"test": MockAction}


class TestAvatarWrapper:
    def setup_method(self):
        global actions_created
        actions_created = []
        self.worker = ConcreteWorker(1, 0)
        self.avatar = avatar_wrapper.AvatarWrapper(
            player_id=None, initial_location=None, avatar_appearance=None
        )

    async def take_turn(self, request_mock):
        with HTTMock(request_mock):

            worker_data = await self.worker.fetch_data(None)
            self.avatar.decide_action(worker_data)

    async def test_bad_action_data_given(self, loop):
        request_mock = InvalidJSONRequest
        await self.take_turn(request_mock)
        assert actions_created == [], "No action should have been applied"

    async def test_non_existant_action(self, loop):
        request_mock = NonExistentActionRequest
        await self.take_turn(request_mock)
        assert actions_created == [], "No action should have been applied"

    def add_effects(self, num=2):
        effects = []
        for _ in range(num):
            effect = MockEffect(self.avatar)
            self.avatar.effects.add(effect)
            effects.append(effect)
        return effects

    def test_effects_on_turn_are_called(self):
        effect1, effect2 = self.add_effects()
        self.avatar.update_effects()
        assert effect1.turns == 1
        assert effect2.turns == 1

    def test_effects_not_removed(self):
        effect1, effect2 = self.add_effects()
        self.avatar.update_effects()
        assert set((effect1, effect2)) == self.avatar.effects

    def test_expired_effects_removed(self):
        effect1, effect2 = self.add_effects()
        effect1.expire = True
        self.avatar.update_effects()
        assert effect2.turns == 1
        assert self.avatar.effects == set((effect2,))

    def test_effects_applied_on_invalid_action(self):
        self.take_turn(InvalidJSONRequest)
        effect = self.add_effects(1)[0]
        self.avatar.update_effects()
        assert effect.turns == 1

    def test_avatar_dies_health(self):
        self.avatar.die(None)
        assert self.avatar.health == 5

    def test_avatar_dies_score_when_large(self):
        self.avatar.score = 10
        self.avatar.die(None)
        assert self.avatar.score == 8

    def test_avatar_dies_score_when_small(self):
        self.avatar.score = 1
        self.avatar.die(None)
        assert self.avatar.score == 0

    def test_avatar_dies_location(self):
        self.avatar.die("test")
        assert self.avatar.location == "test"

    def test_damage_applied(self):
        self.avatar.health = 10
        assert self.avatar.damage(1) == 1
        assert self.avatar.health == 9

    def test_resistance_reduces_damage(self):
        self.avatar.health = 10
        self.avatar.resistance = 3
        assert self.avatar.damage(5) == 2
        assert self.avatar.health == 8

    def test_no_negative_damage(self):
        self.avatar.health = 10
        self.avatar.resistance = 3
        assert self.avatar.damage(1) == 0
        assert self.avatar.health == 10

    def test_calculate_orientation(self):
        # East movement
        self.avatar.location = Location(0, 0)
        self.avatar.previous_location = Location(-1, 0)

        assert self.avatar.calculate_orientation() == "east"



from logging import getLogger
from simulation.direction import Direction
from simulation.event import (
    FailedAttackEvent,
    FailedMoveEvent,
    MovedEvent,
    PerformedAttackEvent,
    ReceivedAttackEvent,
)

LOGGER = getLogger(__name__)


class Action(object):
    def __init__(self, avatar):
        self._avatar = avatar
        try:
            self._target_location = self._avatar.location + self.direction
        except AttributeError:
            self._target_location = self._avatar.location

    @property
    def avatar(self):
        return self._avatar

    @property
    def target_location(self):
        return self._target_location

    def register(self, world_map):
        if world_map.is_on_map(self.target_location):
            world_map.get_cell(self.target_location).actions.append(self)

    def process(self, world_map):
        """Called externally to decide whether to process the action or not."""
        if self._is_legal(world_map):
            self.avatar.previous_location = self.avatar.location
            self._apply(world_map)
            self.avatar.orientation = self.avatar.calculate_orientation()
        else:
            self._reject()

    def _is_legal(self, world_map):
        raise NotImplementedError("Abstract method")

    def _apply(self, world_map):
        raise NotImplementedError("Abstract method")

    def _reject(self):
        raise NotImplementedError("Abstract method")


class WaitAction(Action):
    def __init__(self, avatar):
        super(WaitAction, self).__init__(avatar)

    def _is_legal(self, world_map):
        return True

    def _apply(self, world_map):
        self.avatar.clear_action()


class MoveAction(Action):
    def __init__(self, avatar, direction):
        self.direction = Direction(**direction)
        super(MoveAction, self).__init__(avatar)

    def _is_legal(self, world_map):
        return world_map.can_move_to(self.target_location)

    def process(self, world_map):
        self.detect_cycles(world_map, {self.avatar.location})

    def _apply(self, world_map):
        event = MovedEvent(self.avatar.location, self.target_location)
        self.avatar.add_event(event)

        world_map.get_cell(self.avatar.location).avatar = None
        self.avatar.previous_location = self.avatar.location
        self.avatar.location = self.target_location
        self.avatar.orientation = self.avatar.calculate_orientation()
        world_map.get_cell(self.target_location).avatar = self.avatar
        self.avatar.clear_action()
        return True

    def detect_cycles(self, world_map, visited):
        if not self._is_legal(world_map):
            return self._reject()

        if self.target_location in visited:
            return self._reject()

        next_cell = world_map.get_cell(self.target_location)
        if not next_cell.is_occupied:
            return self._apply(world_map)

        next_action = next_cell.avatar.action
        if next_action.detect_cycles(world_map, visited | {self.target_location}):
            return self._apply(world_map)

        return self._reject()

    def _reject(self):
        event = FailedMoveEvent(self.avatar.location, self.target_location)
        self.avatar.add_event(event)
        self.avatar.clear_action()
        return False


class AttackAction(Action):
    def __init__(self, avatar, direction):
        self.direction = Direction(**direction)
        super(AttackAction, self).__init__(avatar)

    def _is_legal(self, world_map):
        return True if world_map.attackable_avatar(self.target_location) else False

    def _apply(self, world_map):
        attacked_avatar = world_map.attackable_avatar(self.target_location)
        damage_dealt = 1
        self.avatar.add_event(
            PerformedAttackEvent(attacked_avatar, self.target_location, damage_dealt)
        )
        attacked_avatar.add_event(ReceivedAttackEvent(self.avatar, damage_dealt))
        attacked_avatar.damage(damage_dealt)

        self.avatar.clear_action()

        if attacked_avatar.health <= 0:
            # Move responsibility for this to avatar.die() ?
            respawn_location = world_map.get_random_spawn_location()
            attacked_avatar.die(respawn_location)
            world_map.get_cell(self.target_location).avatar = None
            world_map.get_cell(respawn_location).avatar = attacked_avatar

    def _reject(self):
        self.avatar.add_event(FailedAttackEvent(self.target_location))
        self.avatar.clear_action()


ACTIONS = {"attack": AttackAction, "move": MoveAction, "wait": WaitAction}

PRIORITIES = {WaitAction: 0, AttackAction: 1, MoveAction: 2}



from django.contrib.auth.models import User
from django.test import TestCase

from aimmo import models, forms, app_settings

app_settings.GAME_SERVER_URL_FUNCTION = lambda game_id: (
    "base %s" % game_id,
    "path %s" % game_id,
)
app_settings.GAME_SERVER_PORT_FUNCTION = lambda game_id: 0
app_settings.GAME_SERVER_SSL_FLAG = True


class TestForms(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user1 = User.objects.create_user("test", "test@example.com", "password")
        cls.user1.save()
        cls.game = models.Game(id=1, name="test", public=False)
        cls.game.save()

    def test_create_game(self):
        form = forms.AddGameForm([self.game], {"name": "test2"})
        self.assertTrue(form.is_valid())
        new_game = form.save()
        self.assertEqual(new_game.name, "test2")

    def test_cannot_create_duplicate_game(self):
        form = forms.AddGameForm([self.game], {"name": "test"})
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors, {"__all__": [u"Sorry, a game with this name already exists."]}
        )
